// RUN: %parallel-boogie "%s" > "%t"
// RUN: %diff "%s.expect" "%t"

// Permission types
datatype AB { AB(x:int) }
datatype A { A(x:int) }
datatype B { B(x:int) }

// Linear collectors
type {:linear "perm"} Perm = int;

function {:inline} {:linear "perm"} ABCollector(ab: AB) : [int]bool
{ MapConst(false)[ab->x := true][-ab->x := true] }
function {:inline} {:linear "perm"} ABSetCollector(abs: [AB]bool) : [int]bool
{ (lambda i:int :: abs[AB(i)] || abs[AB(-i)]) }

function {:inline} {:linear "perm"} ACollector(a: A) : [int]bool
{ MapConst(false)[a->x := true] }
function {:inline} {:linear "perm"} ASetCollector(as: [A]bool) : [int]bool
{ (lambda i:int :: as[A(i)]) }

function {:inline} {:linear "perm"} BCollector(b: B) : [int]bool
{ MapConst(false)[-b->x := true] }
function {:inline} {:linear "perm"} BSetCollector(bs: [B]bool) : [int]bool
{ (lambda i:int :: bs[B(-i)]) }

function {:inline} bToA (b:B) : A { A(b->x) }

// Global variables
var {:layer 0,1} x:int;
var {:layer 0,1}{:linear "perm"} As:[A]bool;
var {:layer 0,1}{:linear "perm"} Bs:[B]bool;

// Invariant
yield invariant {:layer 1} Inv ();
invariant x >= cardAs(As) - cardBs(Bs);
invariant (forall b:B :: Bs[b] ==> As[bToA(b)]);

yield invariant {:layer 1} Inv_incdec (b:B);
invariant As[bToA(b)];

// Definitions and facts about cardinality
function cardAs (As:[A]bool) : int;
axiom (forall As:[A]bool :: cardAs(As) >= 0);

function cardBs (Bs:[B]bool) : int;
axiom (forall Bs:[B]bool :: cardBs(Bs) >= 0);

pure procedure Lemma_add_to_A (a: A, As: [A]bool);
requires !As[a];
ensures  cardAs(As[a := true]) == cardAs(As) + 1;

pure procedure Lemma_add_to_B (b: B, Bs: [B]bool);
requires !Bs[b];
ensures cardBs(Bs[b := true]) == cardBs(Bs) + 1;

pure procedure Lemma_card_geq (As: [A]bool, Bs: [B]bool);
requires (forall b: B :: Bs[b] ==> As[bToA(b)]);
ensures cardAs(As) >= cardBs(Bs);

pure procedure Lemma_card_greater (_b: B, As: [A]bool, Bs: [B]bool);
requires (forall b: B :: Bs[b] ==> As[bToA(b)]);
requires !Bs[_b];
requires As[bToA(_b)];
ensures cardAs(As) > cardBs(Bs);

// Acutal program
yield procedure {:layer 1} main ({:linear_in "perm"} all_abs: [AB]bool)
requires {:layer 1} all_abs == (lambda v:AB :: 1 <= v->x);
requires call Inv();
{
  var i:int;
  var {:linear "perm"} abs: [AB]bool;
  var {:linear "perm"} ab:AB;
  abs := all_abs;
  i := 1;
  while (*)
  invariant {:yields} true;
  invariant call Inv();
  invariant {:layer 1} 1 <= i;
  invariant {:layer 1} abs == (lambda v:AB :: i <= v->x);
  {
    call ab, abs := transfer_ab(i, abs);
    async call incdec(ab);
    i := i + 1;
  }
}

both action {:layer 1} TRANSFER_AB (x:int, {:linear_in "perm"} abs:[AB]bool) returns ({:linear "perm"} ab:AB, {:linear "perm"} abs':[AB]bool)
{
  assert !abs[AB(-x)];
  assert abs[AB(x)];
  abs' := abs[AB(x) := false];
  ab := AB(x);
}

both action {:layer 1} SPLIT_AB ({:linear_in "perm"} ab:AB) returns ({:linear "perm"} a:A, {:linear "perm"} b:B)
{
  assert ab->x != 0;
  a := A(ab->x);
  b := B(ab->x);
}

yield procedure {:layer 1} incdec({:linear_in "perm"} ab:AB)
requires {:layer 1} ab->x != 0;
preserves call Inv();
{
  var {:linear "perm"} a:A;
  var {:linear "perm"} b:B;
  call a,b := split_ab(ab);
  call {:layer 1} Lemma_card_geq(As, Bs);
  call {:layer 1} Lemma_add_to_A(a, As);
  call geq0_inc(a, b);
  par Inv() | Inv_incdec(b);
  call {:layer 1} Lemma_card_greater(b, As, Bs);
  call {:layer 1} Lemma_add_to_B(b, Bs);
  call geq0_dec(b);
}

atomic action {:layer 1} GEQ0_INC ({:linear_in "perm"} a:A, {:linear "perm"} b:B)
modifies x, As;
{
  assert x >= 0;
  assert !As[a];
  x := x + 1;
  As[a] := true;
}

atomic action {:layer 1} GEQ0_DEC ({:linear_in "perm"} b:B)
modifies x, Bs;
{
  assert x >= 0;
  assert !Bs[b];
  x := x - 1;
  Bs[b] := true;
}

yield procedure {:layer 0} geq0_inc ({:linear_in "perm"} a:A, {:linear "perm"} b:B);
refines GEQ0_INC;

yield procedure {:layer 0} geq0_dec ({:linear_in "perm"} b:B);
refines GEQ0_DEC;

yield procedure {:layer 0} transfer_ab (x:int, {:linear_in "perm"} abs:[AB]bool) returns ({:linear "perm"} ab:AB, {:linear "perm"} abs':[AB]bool);
refines TRANSFER_AB;

yield procedure {:layer 0} split_ab ({:linear_in "perm"} ab:AB) returns ({:linear "perm"} a:A, {:linear "perm"} b:B);
refines SPLIT_AB;
